/*
 * 			cons_impl.h
 *
 * @(#)cons_impl.h 1.17 91/01/11 Copyr 1986 Sun Micro
 *
 * SEE COMMENT AT BEGINNING OF cons_impl.c FOR A DISCUSSION OF THE CODE.
 */

/*
 *	*************** RUNNING OVER 1980 X.25 **********************
 */
#ifndef TRUE
#define TRUE	1
#endif

#ifndef FALSE
#define FALSE	0
#endif

#ifdef OSI_MBUF_TRACE

#define cons_m_get(a)	m_get(a, consmbuf)
#define CONS_MGET(m, i)	MGET(m, i, consmbuf)

#else

#define cons_m_get(a)	m_get(a, MT_DATA)
#define CONS_MGET(m, i)	MGET(m, i, MT_DATA)

#endif

#define CONS_VERSION	1	/* protocol version of this implementation */

/*
 * I believe these single fields are enough to handle the
 * "Target/Lowest-Quality-Acceptable/Available/Selected" handshaking
 * used to determine final values. QOS is currently unimplemented, since
 * 1980 doesn't support it.
 */
struct cons_qos {
	unsigned thru_cg_to_cd;
	unsigned thru_cd_to_cg;
	unsigned transit_delay;
};
typedef struct cons_qos CONS_QOS;

/* 
 * Global variables for this cons instance 
 */
extern struct protosw *cons_x25_proto;

/*
 * X.25 cause codes (not defined in the X.25 code)
 *	From Table 5 of X.25 Protocol Spec. (OSI DIS 8208)
 */

#define DTE_Originated0					0
#define DTE_Originated128				128
#define Number_Busy					1
#define Out_Of_Order					9
#define Remote_Procedure_Error				17
#define Reverse_Charging_Acceptance_Not_Subscribed	25
#define Incompatible_Destination			33
#define Fast_Select_Acceptance_Not_Subscribed		41
#define Ship_Absent					57
#define Invalid_Facility_Request			3
#define Access_Barred					11
#define Local_Procedure_Error				19
#define Network_Congestion				5
#define Not_Obtainable					13
#define RPOA_Out_Of_Order				21
#define Gateway_Detected_Procedure_Error		193
#define Gateway_Congestion				195

/* Ignore high bit, which determines:
 *	0: generated by a public data network
 *	1: generated by a private network
 */
#define X25_CAUSE_TEST(cause) (cause & 0x7f)

typedef enum { /* state name and description */
	STATE_QUIET		= 0,	/* no connection, nothing happenin' */
	STATE_CON_ESTAB		= 1,	/* connection establishment phase   */
	STATE_DATA_XFER		= 2,	/* data transfer phase		    */
	STATE_CON_RELEASE	= 3	/* connection release phase	    */
} CONS_STATE_TYPE;

#define N_STATES	((unsigned) STATE_DATA_XFER + 1)
#define INITIAL_STATE	STATE_QUIET

/*
 * Cons control block. One per connection.
 *
 * Initialization
 *
 *	PROTOSW entries
 *	-- mine [cntl.proto]:  is set during CCB (in cons_hscb_alloc())
 *	-- upper [upper]:      is passed to cons_add_client() and set there
 *			       (currently this defaults to TP0)
 *	-- lower [lower]:      is set during CCB init
 *
 *	PCB entries
 *	-- mine []:	       handle passed up/down from calling layer;
 *	-- upper [cntl.upper]: on CR, passed down from higher layer
 *			       on IND, a 0 is passed up by CONS and this
 *			       field is set by the higher layer
 *	-- lower [cntl.lower]: on IND, passed up from lower layer;
 *			       on CR, a 0 is passed down by CONS and this
 *			       field is set by the lower layer
 *
 * Note: this handle is enough for all interactions. At CR and IND time,
 *	 it will be 0. The passed selectors are used at that time to
 *	 set up the connection. From then on, the handle will be used.
 */
struct ccb {
	PCB_CNTL	cntl;		/* must be first		*/
		/* contains "my" protosw	*/
		/* contains pcb of cons-user	*/
		/* contains pcb of X.25		*/
	struct mbuf	*mbufaddr;	/* mbuf cluster containing this CB */
	XSEL		nsel;		/* i.d. for this CCB	*/
	X25_PID		x25pid;		/* i.d. for this CCB	*/
	CONS_STATE_TYPE state;		/* current state of machine	*/
	struct protosw	*upper_proto;	/* protocol switch to cons-user	*/
	struct protosw	*lower_proto;	/* protocol switch to X.25	*/
	OSI_ADDR	lsap;		/* local SAP			*/
	OSI_ADDR	fsap;		/* foreign SAP			*/
	u_short         version;        /* version number		*/
	CONS_QOS	qos;		/* quality of service 		*/
	u_short		flags;		/* various flags		*/

/*
 * CCB flags
 */
#define X25_CLEAR_REQ_SENT	0x0001	/* sent an x25 clear request out*/

};
typedef struct ccb CCB;

extern CCB	*p_cons_hcb;		/* cons header control block 
					 *  parameters */

#define EACHCCB(p) \
	(p)= (CCB *)p_cons_hcb->cntl.next; (p) != p_cons_hcb; (p)= (CCB *)(p)->cntl.next

/*
 *  Client listener tables.
 *	Transport registers with CONS on a combination of nsel and X25 PID.
 */
struct cons_listener {
	struct cons_listener *next;
	int	hash;		/* hashed cons selector for fast lookup */
	XSEL	nsel;		/* listener address (nsel portion)	*/
	X25_PID	x;		/* listener address (X25 PID portion)	*/
	struct protosw	*proto;	/* client's protosw entry		*/
	int	reference_cnt;	/* number of times listener requested this
				 * address */
};
typedef struct cons_listener CONS_LISTENER;

typedef enum { 		/* originators of disconnects	*/
	ORIG_NS_USER,
	ORIG_NS_PROVIDER,
	ORIG_UNDEFINED
} CONS_ORIGINATOR;

/*
 * CONS DISCONNECT REASONS
 *
 * IMPORTANT: None of the following values (except '0': "Undefined") are
 *	      supported by 1980 X.25.
 *
 * NOTE1: this implementation has separate values for QOS-not-available-perm
 *	  and QOS-not-available-term, depending on whether the originator
 *	  is the NS-user or the NS-provider.
 * NOTE2: "Undefined" in 8878 is taken to mean use '0' ("No Additional
 *	  Information" for the X.25 diagnostic).
 */

typedef enum { 		/* reasons for disconnects	*/
	/*
	 * X.25 VALUES (see Figure 14-B (4 & 5 of 6) in 8208)
	 *
	 *	224-255 are all to support the OSI Network Service Def.
	 *
	 *	0:		Undefined
	 *	224-239:	OSI specific
	 *	240-255:	Higher-level initiated
	 */
		/* "No Additional Info" */
	R_CONS_UNDEFINED = 0,			/* 0 */
		/* originator: NS_PROVIDER */
	R_CP_OSI_NETWORK_SERVICE_PROBLEM = 224,		/* GENERIC */
	R_CP_DIS_TRANS,				/* 225 */
	R_CP_DIS_PERM,				/* 226 */
	R_CP_CON_REJ_UNSPEC_TRANS,		/* 227 */
	R_CP_CON_REJ_UNSPEC_PERM,		/* 228 */
	R_CP_CON_REJ_NO_QOS_TRANS,		/* 229 */
	R_CP_CON_REJ_NO_QOS_PERM,		/* 230 */
	R_CP_CON_REJ_NSAP_UNREACH_TRANS,	/* 231 */
	R_CP_CON_REJ_NSAP_UNREACH_PERM,		/* 232 */
	R_CP_RESET_UNSPEC,			/* 233 */
	R_CP_RESET_CONGESTION,			/* 234 */
	R_CP_CON_REJ_NSAP_UNKNOWN_PERM,		/* 235 */
	R_CP_X25_236,
	R_CP_X25_237,
	R_CP_X25_238,
	R_CP_X25_239,
		/* originator: NS_USER */
	R_CU_HIGHER_LEVEL_INITIATED = 240,		/* GENERIC */
	R_CU_DIS_NORMAL,			/* 241 */
	R_CU_DIS_ABNORMAL,			/* 242 */
	R_CU_X25_243,				/* 243 */
	R_CU_CON_REJ_UNSPEC_TRANS,		/* 244 */
	R_CU_CON_REJ_UNSPEC_PERM,		/* 245 */
	R_CU_CON_REJ_NO_QOS_TRANS,		/* 246 */
	R_CU_CON_REJ_NO_QOS_PERM,		/* 247 */
	R_CU_CON_REJ_INCOMPAT,			/* 248 */
	R_CU_X25_249,				/* 249 */
	R_CU_RESET_USER_RESYNCH,		/* 250 */
	R_CU_X25_251,				/* 251 */
	R_CU_X25_252,				/* 252 */
	R_CU_X25_253,				/* 253 */
	R_CU_X25_254,				/* 254 */
	R_CU_X25_255				/* 255 */
} CONS_REASONS;

/*	##################  MSGSs and EVENTs  ###################	*/

/*
 * All events that CONS has to deal with (CONS will also have to
 * recoginize X.25 type events coming from below).
 *
 * NOTE1: also includes an EVENT to identify CLNS.
 *
 * NOTE2: Expedited Data is not currently implemented.
 *	  Request Confirmation is not currently implemented.
 */
#define CONS_UEVENT	100
#define CONS_PEVENT	200
typedef enum {

	/* User generated events */
	CONS_CONNECT_REQ = CONS_UEVENT,	/* CONNECT request		*/
	CONS_CONNECT_RSP,		/* CONNECT response		*/
	CONS_DISCONNECT_REQ,		/* DISCONNECT request		*/
	CONS_DATA_REQ,			/* DATA request			*/
	CONS_DATA_ACK_REQ,		/* DATA acknowledge request	*/
	CONS_XDATA_REQ,			/* EXPEDITED-DATA request	*/
	CONS_RESET_REQ,			/* RESET request		*/
	CONS_RESET_RSP,			/* RESET response		*/

	/* Provider generated events */
	CONS_CONNECT_IND = CONS_PEVENT,	/* CONNECT indication		*/
	CONS_CONNECT_CNF,		/* CONNECT confirm		*/
	CONS_DISCONNECT_IND,		/* DISCONNECT indication	*/
	CONS_DATA_IND,			/* DATA indication		*/
	CONS_XDATA_IND,			/* XDATA indication		*/
	CONS_DATA_ACK_IND,		/* DATA acknowledge indication	*/
	CONS_RESET_IND,			/* RESET indication		*/
	CONS_RESET_CNF,			/* RESET confirm		*/
	CONS_UPROTO,			/* Protocol violation by
					 * user-generated event		*/
	CONS_PPROTO,			/* Protocol violation by
					 * user-generated event		*/
#define CONS_MAX_PEVENT ((unsigned) CONS_PPROTO)

	/* Other events */
	CLNS_INPUT			/* CLNS INPUT			*/
} CONS_EVENT;

/*
 * Event message definitions
 */

typedef struct {	/* generic message format			*/
	CONS_EVENT	event;		/* message type			*/
} CONS_MSG;

/*
 * Used for connect requests, indications, responses and confirms.
 */
typedef struct {	/* connect service message format		*/
	CONS_EVENT	event;		/* message type			*/
	OSI_ADDR	cd_addr;	/* called address		*/
	OSI_ADDR	cg_addr;	/* calling address		*/
	OSI_ADDR	resp_addr;	/* responding address		*/
	u_char		rcpt_conf;	/* receipt confirmation selection */
	u_char		xdata_sel;	/* expedited data selection	*/
	PROFILE_VECTOR_ARRAY	pv_all;	/* profile vectors		*/
	CONS_QOS	qos;		/* quality of service		*/
	caddr_t		upper_pcb;	/* pcb of higher layer		*/
} CONS_MSG_CONNECT;

/* Note: the responding address in CONS_MSG_DISCONNECT is currently
 *	 unused since X.25 doesn't supply it in the X25_KNL_CLR_IND,
 *	 X25_KNL_CLR_CONF and X25_KNL_RESTART_IND messages (it isn't
 *	part of the 1980 X.25 specification).
 */
typedef struct {	/* connect service message format		*/
	CONS_EVENT	 event;		/* message type			*/
	CONS_ORIGINATOR	 orig;
	CONS_REASONS	 reas;
	OSI_ADDR	 resp_addr;	/* responding address		*/
	caddr_t		 upper_pcb;	/* pcb of higher layer		*/
} CONS_MSG_DISCONNECT;

typedef struct {	/* NS data message format			*/
	CONS_EVENT	event;		/* message type			*/
	u_char		confirm_req;	/* confirmation request		*/
	OSI_ADDR	*cg_osi_addr;	/* calling address		*/
	OSI_ADDR	*cd_osi_addr;	/* called  address		*/
} CONS_MSG_DATA;

typedef struct {	/* NS xdata message format			*/
	CONS_EVENT	event;		/* message type			*/
} CONS_MSG_XDATA;

typedef struct {	/* NS reset message format			*/
	CONS_EVENT	event;	/* message type			*/
	CONS_ORIGINATOR orig;
	CONS_REASONS	reas;
	caddr_t		upper_pcb;	/* pcb of higher layer		*/
} CONS_MSG_RESET;

typedef struct {	/* receive block message format			*/
	CONS_EVENT	event;		/* message type			*/
} CONS_MSG_RCV_BLOCK;

typedef struct {	/* receive unblock message format		*/
	CONS_EVENT	event;		/* message type			*/
} CONS_MSG_RCV_UNBLOCK;

typedef struct {	/* send block message format			*/
	CONS_EVENT	event;		/* message type			*/
} CONS_MSG_SND_BLOCK;

typedef struct {	/* send unblock message format			*/
	CONS_EVENT	event;		/* message type			*/
} CONS_MSG_SND_UNBLOCK;

typedef struct {			/* generic message format	*/ 
	CONS_EVENT      event;          /* message type			*/
	OSI_ADDR	*cg_osi_addr;	/* calling address		*/
	OSI_ADDR	*cd_osi_addr;	/* called  address		*/
} CLNS_MSG_INPUT;
 
/* good for generic declarations */
typedef union {		/* union of all CONS messages */
	CONS_MSG		generic;
	CONS_MSG_CONNECT	connect;
	CONS_MSG_DISCONNECT	disconnect;
	CONS_MSG_DATA		data;
	CONS_MSG_XDATA		xdata;
	CONS_MSG_RESET		reset;
	CLNS_MSG_INPUT		clns;
} CONS_MSG_BUFFER;

/*
 * Message casts for use in cons_action.c
 */
#define CONS_CON(p)		((CONS_MSG_CONNECT *)	  (p))
#define CONS_DIS(p)		((CONS_MSG_DISCONNECT *)  (p))
#define CONS_DAT(p)		((CONS_MSG_DATA *)	  (p))
#define CONS_XDT(p)		((CONS_MSG_XDATA *)	  (p))
#define CONS_RST(p)		((CONS_MSG_RESET *)	  (p))
#define CONS_R_B(p)		((CONS_MSG_RCV_BLOCK *)	  (p))
#define CONS_R_U(p)             ((CONS_MSG_RCV_UNBLOCK *) (p))
#define CONS_S_B(p)             ((CONS_MSG_SND_BLOCK *)	  (p))
#define CONS_S_U(p)             ((CONS_MSG_SND_UNBLOCK *) (p))
