#define _TP_EVENT
/**************************************************************************
 FILE:
	tp_event.h

 DESCRIPTION:
	Events and message headers for communication with the transport
	implementation.

	@(#)tp_event.h 1.18 91/01/11 Copyr 1986 Sun Micro
 **************************************************************************/

/*
 * Parameters.
 */

typedef struct {
	int		dummy;
} TP_QOS_TYPE;

/*
 * Transport (Class 4) events.
 *	Note that (currently) the only connection oriented
 * network events which are used are those necessary to handle
 * the transport connection request necessary to set up a
 * connection for TP0. After that, TP0 takes over.
 */

typedef enum {
	TP_INVALID		= 0,

	/* User-generated events */
	TP_CONNECT_REQ		= 1,
	TP_CONNECT_RESP		= 2,
	TP_DISCONNECT_REQ	= 3,
	TP_DATA_REQ		= 4,
	TP_X_DATA_REQ		= 5,

	/* Provider-generated events */
	TP_CONNECT_IND		= 6,
	TP_CONNECT_CONF		= 7,
	TP_DATA_IND		= 8,
	TP_X_DATA_IND		= 9,
	TP_DISCONNECT_IND	= 10,

#ifdef KERNEL
	TP_RETRANS_TIMER	= 11,
	TP_INACTIVITY_TIMER	= 12,
	TP_REFERENCE_TIMER	= 13,
	TP_CR			= 14,
	TP_CC			= 15,
	TP_ER			= 16,
	TP_DR			= 17,
	TP_DC			= 18,
	TP_EA			= 19,
	TP_DT			= 20,
	TP_AK			= 21,
	TP_ED			= 22,

	/* connection oriented network */
	TP_NCONNECT_CONF	= 23,
	TP_NRESET_IND		= 24,
	TP_NDISCONNECT_IND	= 25,
#endif
} TP_EVENT;

#ifdef KERNEL
/* IMPORTANT: this is also defined (unfortunately) in .../sun/stubs_sunlink.c*/
#define MAX_TP_EVENTS	26
#else KERNEL
#define MAX_TP_EVENTS	11
#endif


/*
 * Stubbed message type for interface declarations.
 */

typedef struct {
	TP_EVENT	tp_event;
} TP_MSG;


/* 
 * Message types to carry parameters with events.  
 * Requests/Responses/Indications/Confirations all handled by single
 * message block.
 */


/* Connect */

typedef struct {
	TP_EVENT		tp_event;
	OSI_ADDR		src_address;
	OSI_ADDR		dst_address;
				/* used for responding address on response */
	u_char			expedited_selected;
	TP_QOS_TYPE		tp_qos;
	PROFILE_VECTOR_ARRAY	pv_all;	/* profile vectors		*/
} TP_MSG_CONNECT;


/* Data */

typedef struct {
	TP_EVENT		tp_event;
	u_char			eot;
} TP_MSG_DATA;


/* Expedited Data */

typedef struct {
	TP_EVENT		tp_event;
} TP_MSG_X_DATA;



/* Disconnect */

typedef enum {

	/* Can be used for all classes */
	DR_UNSPECIFIED		= 0,	/* Reason not specified */
	DR_TSAP_CONGESTION	= 1,	/* Congestion at TSAP */
	DR_TSAP_NOT_CONNECTED	= 2,	/* Session entity not attached to TSAP*/
	DR_ADDRESS_UNKNOWN	= 3,	/* Address unknown */

	/* Classes 1 to 4 */
	DR_NORM_BY_SESSION	= 128,	/* Normal disconnect initiated by
					   session entity */
	DR_REM_TP_CONGESTION	= 128+1,
					/* Remote transport entity congestion
					   at connect request time */
	DR_FAILED_CNCT_NEGOT	= 128+2,
					/* Connection negotiation failed
					   (i.e. proposed class(es) not 
					   supported */
	DR_SRC_REFS		= 128+3,
					/* Duplicate source reference detected
					   for the same pair of NSAP's */
	DR_MISMATCHED_REFS	= 128+4,
					/* Mismatched references */
	DR_PROTOCOL_ERROR	= 128+5,
					/* Protocol error */

	/* 128+6 Not used */


	DR_REF_OVERFLOW		= 128+7,
					/* Reference overflow */
	DR_REFUSED_ON_NET_CNTN	= 128+8,
					/* Connection request refused on this
					   network connection */
	/* 128+9 Not used */

	DR_LENGTH_INVALID	= 128+10,
					/* Header or parameter length invalid*/

	/* Local additions */

	DR_TIMER_EXP		= 128+14, /* TP4 retrans or TS1 */

	/* CONS codes passed up when using TP0 */

	DR_cons			= 256,		/* Base for CONS reson codes */
	DR_C_UNDEFINED  	= DR_cons+0,	/* Undefined */
		/* originator: NS_PROVIDER */
	DR_PROV			= DR_cons+224,	/* GENERIC */
	DR_PROV_DIS_TRANS	= DR_cons+225,	/* Disconnect - transient */
	DR_PROV_DIS_PERM	= DR_cons+226,	/* Disconnect - permanent */
	DR_PROV_REJ_UNSPEC_TRANS= DR_cons+227,	/* Reject - transient */
	DR_PROV_REJ_UNSPEC_PERM = DR_cons+228,	/* Reject - permanent */
	DR_PROV_REJ_NO_QOS_TRANS= DR_cons+229,	/* No QOS - transient */
	DR_PROV_REJ_NO_QOS_PERM = DR_cons+230,	/* No QOS - permanent */
	DR_PROV_REJ_NSAP_UNREACH_TRANS 	= DR_cons+231,	/* NSAP unreachable */
	DR_PROV_REJ_NSAP_UNREACH_PERM  	= DR_cons+232,	/* NSAP unreachable */
	DR_PROV_RESET_UNSPEC	= DR_cons+233,	/* Unspecified RESET */
	DR_PROV_RESET_CONGESTION= DR_cons+234,	/* RESET due to congestion */
	DR_PROV_REJ_NSAP_UNKNOWN_PERM  	= DR_cons+235,	/* Unknown NSAP */
		/* originator: NS_USER */
	DR_USER			= DR_cons+240,	/* GENERIC */
	DR_USER_DIS_NORMAL	= DR_cons+241,	/* Normal disconnect */
	DR_USER_DIS_ABNORMAL	= DR_cons+242,	/* Abnormal disconnect */
	DR_USER_REJ_UNSPEC_TRANS= DR_cons+244,	/* Reject - transient */
	DR_USER_REJ_UNSPEC_PERM	= DR_cons+245,	/* Reject - permanent */
	DR_USER_REJ_NO_QOS_TRANS= DR_cons+246,	/* No QOS - transient */
	DR_USER_REJ_NO_QOS_PERM	= DR_cons+247,	/* No QOS - permanent */
	DR_USER_REJ_INCOMPAT	= DR_cons+248,	/* Incompatable NS user data */
	DR_USER_RESET_USER_RESYNCH = DR_cons+250,	/* User RESET */
} TP_DR_REASON;

/* Error */
typedef enum {
	ER_REASON_UNSPECIFIED	= 0,	/* Reason not specified */
	ER_INVALID_PARAM_CODE	= 1,
	ER_INVALID_TPDU_TYPE	= 2,
	ER_INVALID_PARAM_TYPE	= 3
} TP_ER_REASON;

typedef struct {
	TP_EVENT		tp_event;
	TP_DR_REASON		reason;
} TP_MSG_DISCONNECT;

typedef struct {
	TP_EVENT		tp_event;
	TP_ER_REASON		reject_cause;
} TP_MSG_ERROR;
