/* @(#)tp_impl.h 1.48 90/08/29 Copyr 1986 Sun Micro */

/* Also defined in .../sun/stubs_sunlink.c */
typedef int (*PFI)();

#define T_VERSION		0x01
	/* protocol version */

#define DEF_INITIAL_WINDOW	6
	/* initial receive window size */

#define DEF_MAX_CONNECTIONS	16
	/* maximum TP connections */

#define DEF_TP_MBUF_USAGE	2048
	/* "maximum" mbuf usage by Transport -- i.e. worst-case */

#define DEF_MBUFS_PER_TPQ	((DEF_TP_MBUF_USAGE/DEF_MAX_CONNECTIONS) / 2)
	/* default mbuf usage per transport queue (snd and rcv) */

#define DEF_MAX_RETRANS		6
	/* maximum number of retransmissions */

#define DEF_MAX_PDU_SIZE	1024
#define DEF_PDU_SIZE		128
#define MAX_PDU_SIZE		8192
#define MAX_TP0_PDU_SIZE	2048
#define DEF_DATA_UNIT_SIZ	8192
#define DEF_ED_ON_Q		3

/* 
 * mbuf manipulation.
 */

#define tm_get(a)	m_get(a, MT_DATA)
#define tm_free(a)	m_free(a)
#define tm_freem(a)	m_freem(a)
#define TMGET(m, i)	MGET(m, i, MT_DATA)
#define TMFREE(m, n)	MFREE((m), (n))

struct mbuf *t_params_alloc();

/*
 * State machine.
 */

typedef enum {
	TS_INVALID	= 0,
	TS_REFWAIT	= 1,
	TS_CLOSED	= 2,
	TS_WFCC		= 3,
	TS_WBCL		= 4,
	TS_OPEN		= 5,
	TS_WFTRESP	= 6,
	TS_ACKWAIT	= 7,
	TS_CLOSING	= 8,
	TS_WFNC		= 9
} TSTATE;

/* IMPORTANT: this is also defined (unfortunately) in .../sun/stubs_sunlink.c*/
#define MAX_TP_STATES 	10

#define t_next_state(pcb, state, name)	\
{ \
	(pcb)->t_state = (state); \
	osi_t_state(&osi_tp_dev, PROV_COTS, (caddr_t)pcb, (u_char *)NULL, \
		    (struct mbuf *)NULL, name); \
}

#define t_get_state(pcb)		((pcb)->t_state)

/* Also defined in .../sun/stubs_sunlink.c */
typedef PFI TFSM[MAX_TP_EVENTS][MAX_TP_STATES];

/*
 * Timer manipulations.
 */

#define T_T_TIMES_PER_SECOND	2

/* default values -- in seconds * timer-per-second */

#define DEF_T_VAL_RETRANS		(1 * T_T_TIMES_PER_SECOND)
#define DEF_T_VAL_RETRANS_MAX		(DEF_T_VAL_RETRANS * 64)
					/* Highest backoff value */
#define DEF_T_VAL_WINDOW		(120 * T_T_TIMES_PER_SECOND)
#define DEF_T_VAL_INACTIVITY		(DEF_T_VAL_WINDOW * 8)
#define DEF_T_VAL_REFERENCE		(60 * T_T_TIMES_PER_SECOND)
#define DEF_T_ERROR_WAIT		(10 * T_T_TIMES_PER_SECOND)
					/* Wait after sending ER before
					 * initiating disconnect */
#define DEF_T_CR_TIMEOUT		(60 * T_T_TIMES_PER_SECOND)
#define DEF_T_DR_TIMEOUT		(60 * T_T_TIMES_PER_SECOND)

/* timers */
#define T_T_RETRANS	0
#define T_T_WINDOW	1
#define T_T_INACTIVITY	2
#define T_T_REFERENCE	3
#define T_T_ERROR_SENT	4
#define T_T_TS1		5
#define T_T_TS2		6

#define T_N_TIMERS	7

/*
 * TS1: optional timer; expiration indicates that the connect request
 *	was neither accepted nor refused in the allotted time.
 *	See ISO 8073, section 6.5.4, note 2.
 * TS2: optional timer; expiration indicates that the disconnect request
 *	was not acknowledged in the allotted time.
 *	See ISO 8073, section 6.7.5, note 4.
 *	NOTE: Since TP0 uses the "implicit variant" for disconnection
 *	(i.e. it simply tells the network layer to disconnect instead of
 *	sending a DR TPDU; see ISO 8073 6.7.4), it doesn't use the TS2
 *	timer.
 */

#define t_t_set(pcb, timer, value) \
	((pcb)->timers[(timer)] = value)

#define t_t_stop(pcb, timer) \
	((pcb)->timers[(timer)] = 0)

#define t_t_retranscount_clear(ptcb) \
	(ptcb)->retrans_count = 0;

#define t_t_retranscount_get(ptcb) \
	(ptcb)->retrans_count

#define t_t_retranscount_increment(ptcb) \
	(++((ptcb)->retrans_count))

#define t_t_maxretrans(ptcb) \
	((ptcb)->max_retrans)

/*
 * NB:  initial values are chosen from the HTCB.
 */

/*
 * TBD:	t_t_retrans_set() becomes a function call for dynamaically
 *	adjusted values of retransmission time.
 */

/*
 * If the retransmission timer isn't yet running,
 * t_t_retrans_set() chooses its starting value from the HTCB.
 */ 

#define t_t_retrans_tst(pcb) \
	((pcb)->timers[T_T_RETRANS])

#define t_t_retrans_set(pcb) \
	t_t_set((pcb), T_T_RETRANS, (t_t_retranscount_get(pcb) ? \
		((phtcb->timers[T_T_RETRANS]) * t_t_retranscount_get(pcb)) : \
		(phtcb->timers[T_T_RETRANS])))
		
#define t_t_retrans_stop(pcb) { \
	t_t_stop((pcb), T_T_RETRANS); \
	t_t_retranscount_clear(pcb); \
}

#define t_t_window_set(pcb) { \
	if (pcb->t_flags & TP_FAST_WINDOW) \
		t_t_set((pcb), T_T_WINDOW, \
		    (pcb->fast_window_count == 0 ? phtcb->fast_window_timer : \
			pcb->fast_window_count * phtcb->fast_window_timer)); \
	else \
		t_t_set((pcb), T_T_WINDOW, phtcb->timers[T_T_WINDOW]); \
}

#define t_t_window_stop(pcb) \
	t_t_stop((pcb), T_T_WINDOW)

#define t_t_inactivity_set(pcb) \
	t_t_set((pcb), T_T_INACTIVITY, phtcb->timers[T_T_INACTIVITY])

#define t_t_inactivity_stop(pcb) \
	t_t_stop((pcb), T_T_INACTIVITY)

#define t_t_reference_set(pcb) \
	t_t_set((pcb), T_T_REFERENCE, phtcb->timers[T_T_REFERENCE])

#define t_t_reference_stop(pcb) \
	t_t_stop((pcb), T_T_REFERENCE)

#define t_t_error_sent_set(pcb) \
	t_t_set((pcb), T_T_ERROR_SENT, phtcb->timers[T_T_ERROR_SENT])

#define t_t_error_sent_stop(pcb) \
	t_t_stop((pcb), T_T_ERROR_SENT)

#define t_retrans_initialize(pcb) { \
	t_t_retranscount_clear(pcb); \
        t_t_set((pcb), T_T_RETRANS, (phtcb->timers[T_T_RETRANS])); \
}

#define t_t_ts1_set(pcb) \
	t_t_set((pcb), T_T_TS1, phtcb->timers[T_T_TS1])

#define t_t_ts1_stop(pcb) \
	t_t_stop((pcb), T_T_TS1)

#define t_t_ts2_set(pcb) \
	t_t_set((pcb), T_T_TS2, phtcb->timers[T_T_TS2])

#define t_t_ts2_stop(pcb) \
	t_t_stop((pcb), T_T_TS2)

#define t_t_stop_all_timers(pcb) { \
	t_t_retrans_stop(pcb); \
	t_t_window_stop(pcb); \
	t_t_inactivity_stop(pcb); \
	t_t_reference_stop(pcb); \
	t_t_error_sent_stop(pcb); \
	t_t_ts1_stop(pcb); \
	t_t_ts2_stop(pcb); \
}

/*
 * The Transport Control Block.
 */

struct t_pdu {
	u_short		max_pdu_size;
	u_short		mbufs_per_pdu;
	u_short		pdu_divide;
			/* n right shift to divide by mbufs_per_pdu */
	u_short		rem_ack_delay_time;
	u_short		max_retrans_delay;
	XSEL		cg_tsap_id;
	XSEL		cd_tsap_id;
	u_short		seq_gt_lower_to_ak;
	u_short		loc_ref;
	u_short		rem_ref;
	u_long		max_seq;
	u_long		max_credit;
};


/* Send and receive queues. */

struct t_queue {
	struct mbuf	*first_pdu;
	struct mbuf	*last_pdu;
	struct mbuf	*xda_pdu;

	u_long		seq_lower_edge;
	u_long		seq_upper_edge;

	u_long		seq_in;
			/* next in from upper, or next in from lower */

	u_long		seq_out;
			/* next delivered to upper, or next sent to network */

	u_long		seq_xin;
	u_long		seq_xout;
			/* not used for receiving */

/* ICNCFIX by -TH- change to type int 
	u_long		seq_last_ak_subsequence; */
	int 		seq_last_ak_subsequence;

	u_short		count_mbufs;
			/* for flow control */

	u_short		window;
};

/* The network control block (for CONS) */

struct ncb {
	PCB_CNTL		cntl;
	caddr_t			lower_pcb;
	struct protosw		*lower_proto;
	OSI_ADDR		*osiaddr_dst;
	int			incoming;
	int			can_multiplex;
	int			connected;
	int			ref_cnt;
	T_CLASS			tc;
	int			blocked;
	caddr_t			tcb_p;
};

typedef struct ncb NCB;


/* Performance "hints" for CLNP */
struct ind_hints {
	struct mbuf *mhead;	/* CLNP header - single mbuf */
	int	sp_off;		/* offset of segmentation part */
	int	route_epoch;	/* Used to validate ifp/snpa hints */
	struct ifnet *ifp;
	OSI_DL_ADRS snpa;
	/* for statistics updates */
	u_long		*routecnt;
	u_long		*sourcecnt;
};

	
/* The TCB */

struct tcb {
	PCB_CNTL		cntl;
	caddr_t			lower_pcb;
	caddr_t			upper_pcb;
	struct protosw		*upper_proto;
	struct protosw		*lower_proto;

	OSI_ADDR		osiaddr_src, osiaddr_dst;

	struct mbuf		*mbufaddr; /* mbuf cluster containing this CB */
	TSTATE			t_state;
	u_long	 		t_flags;
	T_CLASS			fsm;
	PROFILE_VECTOR_ARRAY	pv_all;	/* all possible stacks that the
					 * listener will accept calls on
					 */
	PROFILE_VECTOR		pv_one;	/* the active stack (which, in the
					 * case of the listener, will have
					 * been trimmed down from one of
					 * pv_all[]);
					 */
				/* should consolidate the two following
				 * structures into a single union.
				 */
	struct t_queue 		tq_snd, tq_rcv;
	LAYER_FLOW		flow;

	struct t_pdu		pdu;
	u_short			retrans_count;
	u_short			max_retrans;
	u_short			timers[T_N_TIMERS];
	u_short			fast_window_timer;
	u_short			fast_window_count;
	u_long			fast_window_seq;
	u_short			dupl_ak_sent;

	NCB			*pncb;	/* CONS connection CB */
	struct mbuf		*crdata;
	struct mbuf		*discdata;
	TP_DR_REASON		discreason;

	/* The blockedq field is used if CONS is blocked, to queue data
	 * that couldn't be sent.  We look at last_queued_pdu only if the
	 * state is TS_CLOSING - the actions will be different based
	 * on whether it's a DR/DC/ER.  In any other state, the first
	 * thing we do when CONS unblocks us is to empty out the
	 * blocked queue.
	 */ 
	struct mbuf		*blockedq;
	int			last_queued_pdu;
					/* type of latest PDU enqueued on
					 * blockedq
					 */
	struct ind_hints	ind_hints;
};

typedef struct tcb TCB;

/* t_flags */

#define TP_MGMT_PROC		0x00000001	
			/* Transport management process on this TCB */

#define TP_DBG_PROC		0x00000002	
			/* debugging process is on this TCB -- can be same
			  as management process */

#define TP_EXPEDITED_SELECTED	0x00000004
			/* Expedited selected on this connection */

#define TP_CHECKSUMS_ON		0x00000008
			/* Checksums on this connection */

#define TP_EXTENDED_FORMATS	0x00000010
			/* Extended formats on this connection */

#define TP_IS_INITIATOR		0x00000020
			/* This TCB belongs to the Initiator of the connection*/

#define TP_AK_DELAY		0x00000040
			/* Delayed AK has been requested by data input routine*/

#define TP_SND_CREDIT_REDUCED	0x00000080
			/* We are operating in a regime in which the remote 
			   entity has reduced credit */

#define TP_RCV_CREDIT_REDUCED	0x00000100
			/* We are operating in a regime in which we have
			   reduced credit to the remote entity */

#define TP_ERPDU_SENT		0x00000200
			/* An error PDU has been sent -- avoid looping and
			 * resending one for each incoming error.  
			 * Recognized problem:  on consecutive but different
			 * incoming protocol errors, only for the first will
			 * we send an ER PDU.
			 */

#define TP_FLOW_CONTROL		0x00000400
			/* Will use explicit flow control for TP2.
			 */

#define TP_FAST_WINDOW		0x00000800
			/* Will fast window timer for TP4.
			 */

#define TP_REGISTERED		0x00001000
			/* Will fast window timer for TP4.
			 */

#define TP_FASTRETRANS		0x00002000
			/* Has sent one fast retransmit over TP4.
			 * Will not send another fast retransmit until
			 * something is AKed.
			 */

#define CALLED_TSAP_INCLUDED	0x00002000
#define CALLING_TSAP_INCLUDED	0x00004000

/* TCB macros */

#define EACHTCB(p) \
	(p)=(TCB *)phtcb->cntl.next; (p) != phtcb; (p)=(TCB *)(p)->cntl.next

#define EACHNCB(n) \
	(n)=(NCB *)phncb->cntl.next; (n) != phncb; (n)=(NCB *)(n)->cntl.next

/* 
 * Interaction with adjacent layers.
 */

/* ICNCFIX by -TH- for long term ping.  Added last argument to pr_output
   call to match with new specification for ind_output() function call. */

/* NPCTE fix for bugid 4004505, esc 506992. <PP>  <22 October 1996> */
/*
   ind_output() is expecting type u_char for its last parameter: add typecast
   else it may lead to annoying side effects, such as abnormal value for
   the flags in the CLNP packet (data seen as ECHO REQ/RES packets)

   fix compilation bug! The last parenthesis function ended with hints. Put
   it now after 0. It should be noted that the C compiler didn't see the problem
   Thank you mister compiler
*/
/* End of NPCTE fix for bugid 4004505 */

#define TP_SEND(m, ptcb, error)  { \
		CLNS_MSG_INPUT			clns_msg; \
		clns_msg.event = (CONS_EVENT)OSI_NDATREQ; \
		clns_msg.cg_osi_addr = &ptcb->osiaddr_src; \
		clns_msg.cd_osi_addr = &ptcb->osiaddr_dst; \
		osi_t_pdu(&osi_tp_dev, PROV_COTS, (caddr_t)ptcb, \
			  (caddr_t)&clns_msg, m, DIR_OUT); \
		error = ((*ptcb->lower_proto->pr_output)(m, \
			&ptcb->osiaddr_dst, &ptcb->osiaddr_src, \
			(struct mbuf *)NULL, (int *)NULL, &ptcb->ind_hints, \
			(u_char)0 /*-TH-*/) ); \
		osinm_tp_counter(NM_TP_SENT_PDU); \
	}

/*
 * Used to send CONS data.
 */
#define TP_SEND_MSG(m, ptcb, pmsg, error)  {				\
		osi_t_pdu(&osi_tp_dev, PROV_COTS, (caddr_t)ptcb,	\
			(caddr_t)pmsg, m, DIR_OUT);			\
		TP_SEND_MSG_NOW(m, ptcb->pncb, pmsg, error);		\
		osinm_tp_counter(NM_TP_SENT_PDU);			\
	}

/*
 * Used to send CONS control messages if not called form TP_SEND_MSG.
 */
#define TP_SEND_MSG_NOW(m, pncb, pmsg, error)  {			\
		error = (*(pncb)->lower_proto->pr_output)		\
				(m, pncb, &pncb->lower_pcb, pmsg);	\
	}

#ifdef lint
#define tp_receive(m, pmsg, ptcb) \
	((*((ptcb)->upper_proto->pr_input))(m, ptcb, &(ptcb->upper_pcb), \
					    pmsg)) 
#else lint
#define tp_receive(m, pmsg, ptcb) \
( \
	(((&osi_tp_dev)->if_flags & IFF_PROMISC) ? \
		r_osi_t_data(&osi_tp_dev, PROV_COTS, (caddr_t)ptcb, \
			  (caddr_t)pmsg, m, DIR_IN, SDU) : 0), \
	((*((ptcb)->upper_proto->pr_input))(m, ptcb, &(ptcb->upper_pcb), \
					    pmsg)) \
)
#endif not lint

#define tp_ctl_upper(m, pmsg, ptcb) \
	((*((ptcb)->upper_proto->pr_ctlinput)) \
		(PRC_OSI, m, ptcb, &(ptcb->upper_pcb), pmsg))

#define tp_ctl_lower(m, pmsg, ptcb) \
	{ if (ptcb->pncb != NULL) \
		  (*((ptcb)->lower_proto->pr_ctloutput)) \
			  (m, ptcb->pncb, &(ptcb->pncb->lower_pcb), pmsg); }


/* 
 * Addressing and client hashing.
 */

/* addressing */

#define T_CLIENT_TAB_SIZ	37	/* now here's a random distribution */
#define T_TCB_TAB_SIZ		37	/* now here's another one	    */

struct t_client {
	XSEL	 	tsap_id;
	struct protosw 	*proto;
	u_short		ref_cnt;
	T_CLASS		tc;		/* contains classes client reg'd on */
	u_short		hash_value;
			/* extensibility to more complex TSEL's -- 
			don't recompute table entries' TSAP-ID's. */
	struct t_client	*next;
};

/* Hash on only the last byte of the tsel.
 * This is a pretty darn weak hash, but then there
 * currently isn't an overwhelming abundance of tsels
 * in the world. Besides which the numeric ones
 * are liable to be allocated pretty much in ascending order.
 */
#define t_hash_tsap( tsap )					\
	( !tsap.xsel_notnull ? 0 :				\
	  ((tsap.utsel[tsap.xsel_len-1]) % T_CLIENT_TAB_SIZ) )

/* ASSUME that the selectors supplied are from the same layer */

#define t_cmp_tsaps(tsap1, tsap2)					\
		((tsap1.xsel_notnull != 1 && 				\
		  tsap1.xsel_notnull == tsap2.xsel_notnull) ||		\
		 (tsap1.xsel_len == tsap2.xsel_len &&			\
		  !bcmp((char *)tsap1.utsel, (char *)tsap2.utsel, 	\
			(int)tsap1.xsel_len)))
/*
 * PDU manipulations.
 */

/* PDU codes -- also used internally as types */

			/* No 0x00 */
#define ED	0x10	/* Expedited Data */
#define EA	0x20	/* Expedited Data Acknowledgement */
			/* No 0x03 */
			/* No 0x04 */
			/* No 0x05 */
#define AK	0x60	/* Acknowledge */
#define ER	0x70	/* Error PDU */
#define DR	0x80	/* Disconnect Request */
			/* No 0x09 */
			/* No 0xa0 */
			/* No 0xb0 */
#define DC	0xc0	/* Disconnect Confirm */
#define CC	0xd0	/* Connect Confirm */
#define CR	0xe0	/* Connect Request */
#define DT	0xf0	/* Data */

/* maximum user-data on each pdu */
#define CR_UDATA_SIZ	32	/* maximum CR user-data */
#define CC_UDATA_SIZ	32	/* maximum CC user-data */
#define DR_UDATA_SIZ	64	/* maximum DR user-data */
#define DC_UDATA_SIZ	64	/* maximum DC user-data */
#define ED_UDATA_SIZ	16	/* maximum ED user-data */

/* header sizes */
#define MAX_DT_HDR_LEN	12	/* maximum DT header */


/* Parameter identifiers */

#define PAR_DR_ADDL_REASON	0xe0	/* additional info on DR */

#define PAR_PDUSIZE		0xc0	/* maximum TPDU */
#define PAR_CALLING_TSAPID	0xc1	/* calling TSAP-ID */
#define PAR_ERPDU_DUMP		0xc1	/* Dump of PDU that caused ER PDU */
#define PAR_CALLED_TSAPID	0xc2	/* called TSAP-ID */
#define PAR_CHECKSUM 		0xc3	/* checksum */
#define PAR_VERSION		0xc4	/* protocol version */
#define PAR_SECURITY		0xc5	/* protocol version */
#define PAR_ADDL_OPTS		0xc6	/* checksum, expedited data */
#define PAR_PROT_CLASSES	0xc7	/* alternative protocol classes */

#define PAR_ACKTIME		0x85	/* acknowledgement delay */
#define PAR_RESID_ERROR		0x86	/* residual error rate */
#define PAR_PRIORITY		0x87	/* priority */
#define PAR_TRANS_DELAY		0x88	/* transit delay */
#define PAR_THROUGHPUT		0x89	/* throughput */
#define PAR_SUBSEQUENCE		0x8a	/* AK subsequence number */
#define PAR_FLOW_CONTROL	0x8c	/* flow control confirmation */

#define PARSIZ_CHECKSUM		2	/* length of checksum parameter */


#define TP_MBUF_DIVIDE		7	/* divide max_pdu_size by to get
					 * mbufs per TPDU */

#define ACK_DELAY_TIME		250	/* msec, fasttimeout value */

/*
 * For passing parameters between caller and PDU encode/decode routines,
 * and for storing parameters with a PDU on a queue.
 * NB: types are ordered for alignment in arbitrary buffer (e.g. mbuf).
 */
struct tpdu_params {
	u_char			pdu_type;

	/* DT */
	u_char			eot;
	u_short			count_mbufs;

	/* General purpose */
	u_short			loc_ref;
	u_short			rem_ref;
	u_long			pdu_nr;
	u_long			credit;
	u_long			seq_to_ack;
	u_long			pdu_len;
	u_long			header_len;

	/* AK and EA */
	u_short			subsequence_number;
	u_short			your_subsequence_number;
	u_long			your_pdu_nr;
	u_short			your_credit;
	u_char			flow_control_confirm;

	/* CR and CC */
	T_CLASS			tc;
	u_char			version;
	u_char			extended_formats;
	u_char			checksums_on;
	u_char			expedited_requested;
	u_char			flow_control;
	u_short			max_pdu_size;
	u_short			ack_delay_time;

	XSEL			*calling_tsap_id;
	XSEL			*called_tsap_id;
	u_char			calling_tsap_included;
	u_char			called_tsap_included;
				/* latter two are booleans:  tsap on PDU? */

	/* DR and ER */
	u_char			reason;
};


/*
 * Manipulations of sequence numbers (in a tpdu_params structure),
 * and delayed acknowledgement.
 */
#define MAX_7BIT_SEQ	0x0000007f
#define MAX_32BIT_SEQ	0x7fffffff

/*
 * n1 < n2 < n3?
 */
#define t_seq_between(n1, n2, n3) \
	(((n1) <= (n2) && (n2) < (n3)) || \
	 ((n3) < (n1) && ((n1) <= (n2) || (n2) < (n3))))

/*
 * n1 > n2 ?
 * NB: For sequence numbers within window.
 */
#define t_q_seq_gt(tq, n1, n2) \
	(((n1) > (n2)) ? \
		((tq).seq_lower_edge <= (n2) || (tq).seq_lower_edge > (n1)) : \
	(((n1) < (n2)) ? \
		((tq).seq_lower_edge <= (n2) && (tq).seq_lower_edge > (n1)) : \
		FALSE))

/*
 * result = previous + 1.
 */

#define t_seq_assign_next(ptcb, previous) \
		(((previous) + 1) & (ptcb)->pdu.max_seq)

/*
 * Increment sequence number.
 */

#define t_seq_inc(ptcb, seq) \
		(seq) = t_seq_assign_next(ptcb, seq)

/*
 * result = previous + n.
 */

#define t_seq_assign_nmore(ptcb, from, nmore) \
		(((from) + (nmore)) & (ptcb)->pdu.max_seq)

/*
 * Given upper and lower edges, calculate the credit.
 * credit = upper - lower
 */

#define t_seq_get_credit(ptcb, lower, upper) \
	(((upper) > (lower)) ? ((upper) - (lower)) : \
	(((upper) == (lower)) ? 0 : \
		((ptcb->pdu.max_seq - (lower)) + (upper) + 1)))

/*
 * Given higher and lower values, get the difference.
 * diff = higher - lower
 */
#define t_seq_diff(ptcb, higher, lower) \
		t_seq_get_credit(ptcb, lower, higher)

/*
 * If the last sequence number received is seq_gt_lower_to_ak greater than
 * the lower window edge (seq_lower_edge), AK immediately.  Also, if the
 * difference between the last sequence number received and the upper edge
 * is very low (2 for now), send AK immediately.  Otherwise, set up for
 * delayed AK.
 */
#define t_doAK(ptcb) { \
	if (t_seq_diff(ptcb, (ptcb)->tq_rcv.seq_in, \
	     (ptcb)->tq_rcv.seq_lower_edge) >= (ptcb)->pdu.seq_gt_lower_to_ak \
	 || t_seq_diff(ptcb, (ptcb)->tq_rcv.seq_upper_edge, \
	     (ptcb)->tq_rcv.seq_in) <= 2) { \
		(void) t_sendAK(ptcb); \
	} else \
		(ptcb)->t_flags |= TP_AK_DELAY; \
}



/*
 * mbuf manipulations
 */

#define t_prepend_mbuf(m, n) { \
		m->m_next = n; \
	}

#define t_prepend_chain(m, n) { \
		register struct mbuf *mtmp1; \
		mtmp1 = (m); \
		while (mtmp1->m_next) mtmp1 = mtmp1->m_next; \
		mtmp1->m_next = (struct mbuf *)(n); \
	}

#define t_seqget(m)	(mtod((m), struct tpdu_params *))->pdu_nr

#define t_sum_cc(m, i) { \
		register struct mbuf *mtmp1; \
		mtmp1 = (m); (i) = 0; \
		while (mtmp1) { (i) += mtmp1->m_len; mtmp1 = mtmp1->m_next; } \
	}
	

#define t_sum_mb(m, i) { \
		register struct mbuf *mtmp1; \
		mtmp1 = (m); (i) = 0; \
		while (mtmp1) { \
			if (mtmp1->m_len > MLEN) { \
				if (mtmp1->m_cltype == 1) (i) += OSI_CL1_MBUF; \
				else (i) += OSI_CL2_MBUF; \
			} else ++(i); \
			mtmp1 = mtmp1->m_next; \
		} \
	}
			
#define t_dataq_append(first, last, pel) { \
		if (!(first)) \
			(first) = (last) = (pel); \
		else { \
			if (!(last)) \
				panic("t_dataq_append"); \
			(last)->m_act = (pel); \
			(last) = (pel); \
		} \
		(pel)->m_act = NULL; \
	}

#define t_anyq_append(first, pel) { \
		register struct mbuf *mtmp1; \
		if ((first) == NULL) (first) = (pel); \
		else { \
			mtmp1 = (first); \
			while (mtmp1->m_act) mtmp1 = mtmp1->m_act; \
			mtmp1->m_act = (pel); \
		}\
		(pel)->m_act = NULL; \
	}
			
/*
 * This macro assumes that the first mbuf is a pointer to a params
 * structure -- it has to delete the mbuf allocated for call{ing,ed}_tsap_id
 * before deleting the rest.  Delete all elements upto and including "to".
 */
#define t_anyq_delete(head, to) {					\
	register struct mbuf *mtmp1, *mtmp2;				\
	mtmp1 = (head);							\
	while (mtmp1) {							\
		register struct tpdu_params *pparams2;			\
		pparams2 = mtod(mtmp1, struct tpdu_params *);		\
		if (pparams2->calling_tsap_id)				\
			tm_free(dtom(pparams2->calling_tsap_id));	\
		if (mtmp1 == (to)) {					\
			(head) = (to)->m_act;				\
			tm_freem(mtmp1);				\
			break;						\
		} else {						\
			mtmp2 = mtmp1;					\
			mtmp1 = mtmp1->m_act;				\
			tm_freem(mtmp2);				\
		}							\
	}								\
}

/*
 * NB:	Takes into account the header mbuf with the tpdu_params.
 */
#define t_anyq_copy(from) \
		m_copy((from)->m_next, 0, M_COPYALL)

/*
 * NB: If (to) is the last element, clear tail pointer.
 */
#define t_dataq_delete(head, tail, to) {				\
	if ((to) && !(to)->m_act) (tail) = NULL;			\
	t_anyq_delete((head), (to));					\
}

/*
 * Set pafter to 0 to insert at head of queue.
 */
#ifdef lint
#define t_anyq_insert(pfirst, pafter, pel) {}
#else lint
#define t_anyq_insert(pfirst, pafter, pel) { \
		if (pafter) { \
			(pel)->m_act = (pafter)->m_act; \
			(pafter)->m_act = (pel); \
		} else { \
			(pel)->m_act = (pfirst); \
			(pfirst) = (pel); \
		} \
	}
#endif lint

#define t_dataq_insert(pfirst, plast, phere, pel) { \
		if ((phere) == (plast)) (plast) = (pel); \
		t_anyq_insert(pfirst, phere, pel); \
	}
		

/*
 * Associated with parameters available to and configurable by network 
 * management.
 */

#define TP_WRITE_PARAMS		1
#define TP_READ_PARAMS		2
#define TP_MAX_AK_TIME		200
				/* Maximum AK delay -- assume worst case and
				 * delayed acknowledgement */

/* 
 * PDU sizes, organized by (PDU parameter code, size) .
 * NB:  Must be organized by smallest first !! 
 */

#define N_CHOICES_PDU_SIZ	7 /* Number of size choices */
extern u_short pdusiz[N_CHOICES_PDU_SIZ][2];


/*
 * Routine to build/send AK without FCC
 */
#define t_buildAK(ptcb)	t_build_fc_AK(ptcb, (struct tpdu_params *)NULL)
#define t_sendAK(ptcb)	t_send_fc_AK(ptcb, (struct tpdu_params *)NULL)

/*
 * Check if it's TP2 with flow control.
 */
#define tp2_flow_control(ptcb)	((ptcb->pv_one & T_CLASS_MASK) == T_CLASS2 && \
					(ptcb->t_flags & TP_FLOW_CONTROL))

/*
 * Macro to piggy-back AK on a DT PDU.  Used by TP4/TP2.
 * Note: AK PDU is contained within one small mbuf.
 */
extern int tp_piggybacked_acks;
#define PIGGY_BACK_AK(m, ptcb) { \
	if (ptcb->t_flags & TP_AK_DELAY && tp_piggybacked_acks) { \
		struct mbuf *mtmp1; \
		if ((mtmp1 = t_buildAK(ptcb)) != NULL) { \
			t_prepend_mbuf(mtmp1, m); \
			m = mtmp1; \
		} \
	} \
}
